# #############################################################################
#  This file is part of df_config                                             #
#                                                                             #
#  Copyright (C) 2020 Matthieu Gallet <df_config@19pouces.net>                #
#  All Rights Reserved                                                        #
#                                                                             #
#  You may use, distribute and modify this code under the                     #
#  terms of the (BSD-like) CeCILL-B license.                                  #
#                                                                             #
#  You should have received a copy of the CeCILL-B license with               #
#  this file. If not, please visit:                                           #
#  https://cecill.info/licences/Licence_CeCILL-B_V1-en.txt (English)          #
#  or https://cecill.info/licences/Licence_CeCILL-B_V1-fr.txt (French)        #
#                                                                             #
# #############################################################################
"""__init__ file."""
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("df_config")
except PackageNotFoundError:
    __version__ = "1.0.0"
